import numpy as np
import matplotlib.pyplot as plt
import tqdm
import time

class GP:
    def __init__(self, K, gamma):
        self.gamma = gamma
        self.mu = 0
        self.var = 0
        self.cov_ = None
        self.inv_cov_ = None
        self.samples = []
        self.y = []
        self.K = K
    
    def update(self, i, y):
        '''
        Update the posterior of this GP.
        Note that this retains the covariance from scratch 
        - which is feasible for small amounts of samples but has t^4 complexity after t samples
        - if we do schur complement - then t^3
        - however if used with weights, I think we want to retrain
        - size of weights should match len(samples)
        '''
        self.samples.append(i)
        self.y.append(y)
        samples = self.samples        
        self.cov_ = self.K[np.ix_(samples, samples)]+self.gamma*np.eye(len(samples))
        self.inv_cov_ = np.linalg.inv(self.cov_)
        # k_t(x), pull out the relevant part of the kernel matrix |X| x samples
        self.Ks = self.K[:,samples] 
        # k_t(x)(K_t + sigma I)^{-1}, kernelize the points - |X| x samples
        self.A = self.Ks@self.inv_cov_     
        self.mu = self.A@np.array(self.y)
        self.var = (np.diag(self.K) 
                    - np.squeeze(self.A[:,np.newaxis, :]@self.Ks[:,:,np.newaxis]))

    def debug(self, phiX):
        cov = self.gamma*np.eye(phiX.shape[1])
        running_sum = 0
        for i in range(len(self.samples)):
            running_sum += phiX[self.samples[i]] * self.y[i]
            cov += np.outer(phiX[self.samples[i]],phiX[self.samples[i]])
        inv_cov = np.linalg.inv(cov)
        print('mu', phiX@inv_cov@running_sum)
        print('gp.mu', self.mu)
        print((phiX@inv_cov)[:,:,np.newaxis].shape, phiX[:,np.newaxis,:].shape)
        print('var', np.squeeze((phiX@inv_cov)[:,np.newaxis,:]@phiX[:,:,np.newaxis])*self.gamma)
        print('gp.var', self.var)

        
class GP_for_schur:
    def __init__(self, K, gamma):
        self.gamma = gamma
        self.mu = 0
        self.var = 0
        self.cov_ = None
        self.inv_cov_ = None
        self.samples = []
        self.y = []
        self.K = K
    
    @staticmethod
    def schur_inverse(K, Kinv, x, var):
        '''
        Use Schur's complement to compute the inverse of symmetric 
        matrix: new_cov_inv = [[K, x], [x^T, var]] given the inverse of K. 
        K is [t x t] array. x is a t x 1 vector. var is a non_zero scalar. 
        '''
        # precompute some things
        tmp = Kinv @ x
        tmp_dotx = tmp.dot(x)
        a = var - tmp_dotx
        b = tmp_dotx / a
        c = (1 + b) / var
        # allocate memory for new matrix: 
        t = K.shape[0]
        new_inv = np.zeros((t+1, t+1))
        # use Schur's complement to compute inverse
        new_inv[:t, :t] = Kinv + np.outer(tmp, tmp) / a
        new_inv[:t, -1] = new_inv[-1, :t] = -c * tmp
        new_inv[-1, -1] = (1 + c * tmp_dotx) / var
        return new_inv 

    def update_cov_and_inv(self):
        '''
        Update covariance after new sample taken
        '''    
        self.cov_ = self.K[np.ix_(self.samples, self.samples)]+self.gamma*np.eye(len(self.samples))
        if len(self.samples) == 1:
            self.inv_cov_ = np.linalg.inv(self.cov_)
        else:
            self.inv_cov_ = GP_for_schur.schur_inverse(self.cov_[:-1, :-1],   # first t samples take
                                                       self.inv_cov_,      # old inverse, not yet updated
                                                       self.cov_[-1, :-1],  # new off diagonal term
                                                       self.cov_[-1, -1])  # new diagonal term

    def update(self, i, y):
        '''
        Update the posterior of this GP.
        Note that this retains the covariance from scratch 
        - which is feasible for small amounts of samples but has t^4 complexity after t samples
        - if we do schur complement - then t^3
        - need to replace it with schur complement and using the kernel directly
        '''
        # record new sample drawn
        self.samples.append(i)
        self.y.append(y)
        # update covariance and inverse covariances
        self.update_cov_and_inv()
        # k_t(x), pull out the relevant part of the kernel matrix |X| x samples
        self.Ks = self.K[:,self.samples] 
        # k_t(x)(K_t + sigma I)^{-1}, kernelize the points - |X| x samples
        self.A = self.Ks@self.inv_cov_     
        self.mu = self.A@np.array(self.y)
        self.var = np.diag(self.K) - np.squeeze(self.A[:,np.newaxis, :]@self.Ks[:,:,np.newaxis])

        
def rhogp(Y, K, num_iters, step_size = 1, gamma = .1, debug=False, viz_step=10):
    '''
    Y are subsets of indices
    Y needs to be a subset of X, and K should be a kernel on X
    '''
    n = K.shape[0]
    samples = [] # maintain the indices of X's seen
    l = np.zeros(n)    
    idx = np.random.randint(n)
    l[idx] = 1
    samples.append(idx)
    grad_norms = []
    for t in range(1,num_iters):
        # find element with highest variance
        w = np.sqrt(l[samples])
        cov_ = K[np.ix_(samples, samples)]*np.outer(w,w)+gamma*np.eye(len(samples))
        inv_cov_ = np.linalg.inv(cov_)
        Ks = K[:,samples]*w # pull out the relevant part of the kernel matrix |X| x samples
        A = Ks@inv_cov_ # kernelize the points - |X| x samples
        var = np.diag(K)/gamma - np.squeeze(A[:,np.newaxis, :]@Ks[:,:,np.newaxis])/gamma        
        y_idx = Y[np.argmax(var[Y])] 
        
        # now compute the gradient at this point
        g = K[y_idx] - A[y_idx]@Ks.T
        g = - g*g/(gamma*gamma)
        if debug:
            print('Y', Y)
            print('variance', var)
            print('y_idx', y_idx, 'g', g, 'gamma', gamma)
            print('samples', samples)
            print('l', l)
            print('max', max(var[Y]), y_idx,'\n')
        grad_norms.append(np.linalg.norm(g - np.sum(g)/n*np.ones(n)))
        if t % (viz_step) == 0:
            fig, ax = plt.subplots(1,2)
            ax[0].plot(grad_norms)
            ax[1].plot(l)
            print('max', max(var[Y]), idx,'\n')
            plt.show()
        # decide on the direction
        idx = np.argmin(g)
        eta = step_size/(t+2)
        lold = l
        l = (1-eta)*l+eta*np.eye(n)[idx]
        samples = [i  for i in range(n) if l[i]>1e-6]
    return lold, max(var[Y])   

def debug_rhogp(l, yidx, gamma, X, K):
    n = len(X)
    d = X.shape[1]
    # line to check variances
    A = gamma*np.eye(d)
    for i in range(len(X)):
        A += l[i]*np.outer(X[i],X[i])
    A = np.linalg.inv(A)
    print('var', [X[i]@A@X[i] for i in range(n)])
    # line to check gradient
    print('g', [-(X[yidx]@A@X[i])**2 for i in range(n)])

if __name__ == '__main__':
    np.random.seed(10)
    X = np.random.randn(10,5)
    K = X@X.T
    gp = GP(K, .1)
    print('1 sample test')
    gp.update(3, .7)
    idx = 5
    print('var test', K[idx,idx] - K[idx,3]*K[idx,3]*(K[3,3]+.1)**(-1), gp.var[idx])
    print('mu test', K[idx,3]*(K[3,3]+.1)**(-1)*.7, gp.mu[idx])
    print('many sample test')
    for i in range(10):
        gp.update(np.random.randint(10), np.random.randn())
    gp.debug(X)
    
    
