import numpy as np
from numpy import sqrt, log, exp
import itertools
import logging
import time
from library import *
import matplotlib.pyplot as plt 
import GP
logging.basicConfig(level= "INFO")

class MELK(object):
    def __init__(self, means, K, tau, 
                 noise_sig,gamma=1,
                 conf_const = 2.0, 
                 max_samples=100,  batch_size=100,
                 verbose=False):
        self.K = K
        self.M = K.shape[0]
        self.tau = tau
        self.means = means
        self.arm_count = np.zeros(self.M)
        self.noise_sig = noise_sig      
        self.max_samples = max_samples
        self.arm_counts = np.zeros(self.M)
        self.true_set = [i for i in range(self.M) if self.means[i] > self.tau]
        self.f1_scores = []
        self.precision = []
        self.recall = []
        self.batch_size = batch_size    
        self.running_sum = 0
        self.gamma = gamma
        self.active_arms = np.arange(self.M, dtype=np.int32)
        self.good_arms = []
        self.bad_arms = []
        self.beta = conf_const
        self.sample_points = []
        self.verbose = verbose
        
    def algorithm(self, step_size=1, design_iters=1000, viz_step = 10):   
        print('noise_sig', self.noise_sig, 
              'self.beta', self.beta, 
              'self.tau', self.tau, 'self.gamma', self.gamma)
        self.N = 0
        self.phase_index = 1
        self.gp = GP.GP_for_schur(self.K, self.noise_sig**2)
        while len(self.active_arms) > 0 and self.N < self.max_samples:                
            logging.info('about to compute allocation')
            # Computing the design and getting estimates
            design, rho = GP.rhogp(self.active_arms, self.K, num_iters=design_iters, 
                                   step_size=step_size, gamma=self.gamma/(self.N+1), viz_step=viz_step)
            num_samples = self.batch_size
            sampled_arms = np.random.choice(self.M, size=num_samples, p=design) 
            for idx in sampled_arms:
                self.arm_count[idx] += 1
                y = self.means[idx] + np.random.randn() * self.noise_sig
                self.gp.update(idx,  y)    
            self.N += num_samples
            self.sample_points.append(self.N)
            self.upper = self.gp.mu + np.sqrt(self.gp.var) * self.beta 
            self.lower = self.gp.mu - np.sqrt(self.gp.var) * self.beta 
            self.update()
            self.update_metrics()
            if self.verbose == True:
                fig, ax = plt.subplots(1,2)
                ax[0].plot(self.upper, label = "up")
                ax[0].plot(self.lower, label = "dw")
                ax[0].plot(self.means, c = "g", label = "m")
                ax[0].plot(self.gp.mu, 'o', c = "g", label = "em")
                ax[1].plot(self.arm_count)
                plt.show()
            
            logging.info('\n\n')
            logging.info('finished phase {}'.format(self.phase_index))
            logging.info('round sample count {} total sample count {}'.format(self.N,
                                                                              self.batch_size))
            logging.info('active_set {}'.format(len(self.active_arms))) 
            logging.info('{} arms declared good {}'.format(len(self.good_arms),str(self.good_arms[:10]))) 
            logging.info('{} arms declared bad {}'.format(len(self.bad_arms),str(self.bad_arms[:10]))) 
            logging.info('precision:{}, recall: {}'.format(self.precision[-1], self.recall[-1])) 
            logging.info('lengths %s' %str(len(self.upper)))
            self.phase_index += 1            

    def update(self):         
        keep_indices = []        
        for arm_idx in self.active_arms:
            if self.tau > self.upper[arm_idx]: 
                self.bad_arms.append(arm_idx)
            elif self.tau < self.lower[arm_idx]:
                self.good_arms.append(arm_idx)
            else:
                keep_indices.append(arm_idx)
        self.active_arms = keep_indices

    def update_metrics(self):
        p = compute_precision(self.good_arms, self.true_set)
        r = compute_recall(self.good_arms, self.true_set)
        self.precision.append(p)
        self.recall.append(r)
        if p == 0 and r == 0: f1 = 0
        else: f1 = 2 * p * r / (p + r)
        self.f1_scores.append(f1)