import numpy as np 
from scipy.sparse.linalg import eigsh
import matplotlib.pyplot as plt
from library import compute_precision, compute_recall
import GP
import time

class TRUVAR:
    def __init__(self, means, K, tau, 
                 noise_sig, conf_const=1.96,
                 max_samples=200, record_each=1000,
                 verbose=False, seed = None):
        np.random.seed(seed)
        self.K = K
        self.n = self.K.shape[0]
        self.noise_sig = noise_sig      # std dev of noise
        self.verbose = verbose
        self.G = []     # good arms
        self.B = []      # bad arms
        self.tau = tau      # threshold
        self.sampled = []   # previously sampled arms
        self.means = means
        self.max_samples = max_samples
        self.arm_count = np.zeros(self.n) # Tracks individual arm pulls
        self.f1_scores = []
        self.precision = []
        self.recall = []
        self.true_set = [i for i in range(self.n) if self.means[i] > self.tau]
        self.sample_points = []
        self.record_each = record_each
        self.beta = conf_const
        self.iter_time = 0
        
    def pull_arm(self, i):
        y = self.means[i] + self.noise_sig * np.random.randn()
        self.gp.update(i, y)
        self.sampled.append(i)
        self.arm_count[i] += 1    # Count arm pull
        self.update_bounds()
    
    def update_bounds(self):
        self.upper = self.gp.mu + self.beta * np.sqrt(self.gp.var)
        self.lower = self.gp.mu - self.beta * np.sqrt(self.gp.var)
        
    def update_good_bad_sets(self):
        for i in range(self.n):
            if i not in self.G and i not in self.B:
                if self.lower[i] > self.tau: self.G.append(i)
                elif self.upper[i] < self.tau: self.B.append(i)

    def select_arm(self):
        idxs = [i  for i in range(self.n) if ((i not in self.G) and (i not in self.B))]
        K_active = self.K[idxs, :] #active part of kernel |A|x|X|
        A_active = self.gp.A[idxs, :]
        # K_active - |A|x |X|,  A_active - |A|xsamples, gp.Ks[:,:,np.newaxis]- samples x |X|
        cov = K_active - A_active@self.gp.Ks.T
        #cov.T@cov - |X|x|X|
        a = np.diag(cov.T@cov)/(self.gp.var+self.noise_sig**2)
        return np.argmax(a)
    
    def update_metrics(self):
        p = compute_precision(self.G, self.true_set)
        r = compute_recall(self.G, self.true_set)
        self.precision.append(p)
        self.recall.append(r)
        if p == 0 and r == 0: f1 = 0
        else: f1 = 2 * p * r / (p + r)
        self.f1_scores.append(f1)

    def run(self):
        '''
        Run LSE
        '''
        # choose random starting index
        self.gp = GP.GP_for_schur(self.K, self.noise_sig**2)        
        start = np.random.randint(0,self.n-1)
        self.pull_arm(start)
        while len(self.G) + len(self.B) < self.n \
            and len(self.sampled) < self.max_samples:
            arm = self.select_arm()
            self.pull_arm(arm)
            self.update_good_bad_sets()
            t = len(self.sampled)
            if not t % self.record_each: 
                self.update_metrics()
                self.sample_points.append(t)
                if self.verbose:
                    print('Samples {}'.format(t))
                    print('{} arms declared good {}'.format(len(self.G), 
                                                            sorted(self.G[:10])))
                    print('{} arms declared bad {}'.format(len(self.B), 
                                                           sorted(self.B[:10])))
                    print('{} arms remaining'.format(self.n - len(self.G) - len(self.B)))
                    print('precision:{}, recall: {}'.format(self.precision[-1],
                                                            self.recall[-1])) 
                    print('{} average iter time'.format(self.iter_time/t))
                    fig, ax = plt.subplots(1,2)
                    ax[0].plot(self.upper, label = "up")
                    ax[0].plot(self.lower, label = "dw")
                    ax[0].plot(self.means, c = "g", label = "m")
                    ax[0].plot(self.gp.mu,'o', c = "g", label = "em")
                    ax[1].plot(self.arm_count)
                    ax[0].legend()
                    plt.show()






