import os
os.environ["OMP_NUM_THREADS"] = "2" # export OMP_NUM_THREADS=4
os.environ["OPENBLAS_NUM_THREADS"] = "2" # export OPENBLAS_NUM_THREADS=4 
os.environ["MKL_NUM_THREADS"] = "4" # export MKL_NUM_THREADS=6
os.environ["VECLIB_MAXIMUM_THREADS"] = "2" # export VECLIB_MAXIMUM_THREADS=4
os.environ["NUMEXPR_NUM_THREADS"] = "2" # export NUMEXPR_NUM_THREADS=6
from optparse import OptionParser
import numpy as np
import numexpr as ne 
from multiprocessing import RawArray, Pool
from LSEGP import LSE
from TRUVARGP import TRUVAR
from MELKGP import MELK
import pickle
from numpy import sin, cos


def parse_argv():
    # set up parser
    parser = OptionParser()
    parser.add_option("--d", default=1, action="store", dest="d", type="int")
    parser.add_option("--n", default=700, action="store", dest="n", type="int")
    parser.add_option("--sig", default=.2, action="store", dest="sig", type="float")
    parser.add_option("--noise_sig", default=.1, action="store", dest="noise_sig", type="float")
    parser.add_option('--num_trials', default=25, action="store", dest="num_trials", type="int")
    parser.add_option("--max_samples", default=100, action="store", dest="max_samples", type="int")
    parser.add_option("--design_iters", default=500, action="store", dest="design_iters", type="int")
    parser.add_option("--conf_const", default=1.96, action="store", dest="conf_const", type="float")
    parser.add_option("--batch_size", default=5, action="store", dest="batch_size", type="int")
    parser.add_option("--frac", default=0.3, action="store", dest="frac", type="int")
    parser.add_option("--gamma", default=.1, action="store", dest="gamma", type="float")
    parser.add_option("--verbose", action="store_true", dest="verbose")
    options, _ = parser.parse_args()
    # save to dictionary
    global var_dict  # GLOBAL dictionary of parameters
    var_dict = vars(options)

def parallel_kernel(X, gamma):
    if len(X.shape) == 1: X = np.expand_dims(X, axis=1)
    X_norm = np.einsum('ij,ij->i',X,X)
    K = ne.evaluate('exp(-g * (A + B - 2 * C))', {
                     'A' : X_norm[:,None],
                     'B' : X_norm[None,:],
                     'C' : np.dot(X, X.T),
                     'g' : gamma
    })
    return K

########## Build environments ############
def build_env1():
    n = var_dict['n']
    sig = var_dict['sig']
    gamma = 1/ (2 * sig**2)
    X = np.linspace(0, 1, n)
    K = parallel_kernel(X, gamma)
    return shareable_array(K)

def build_env1_means():
    n = var_dict['n']
    sig = var_dict['sig']
    K = var_dict['K']
    K = np.frombuffer(var_dict['K']).reshape(var_dict['K_shape'])
    means = np.random.multivariate_normal(np.zeros(n), K)
    t = int(np.ceil(n*var_dict['frac']))
    tau = sorted(means, reverse = True)[t]
    return means, tau


def build_env2():
    n = var_dict['n']
    sig = var_dict['sig']
    gamma = 1/ (2 * sig**2)
    X = np.linspace(0, 1, n)
    K = parallel_kernel(X, gamma)
    return shareable_array(K)

def build_env2_means():
    n = var_dict['n']
    sig = var_dict['sig']
    means = means = [np.cos(8*np.pi/n*i) for i in range(n)]
    t = int(np.ceil(n*var_dict['frac']))
    tau = sorted(means, reverse = True)[t]
    return means, tau



def build_env3(seed = None):
    sig = var_dict['sig']
    np.random.seed(seed)
    x_grid = 30
    y_grid = 30
    x = np.linspace(-1, 1, x_grid)
    y = np.linspace(-1, 1, y_grid)
    X1, Y1 = np.meshgrid(x, y)
    X = np.concatenate((X1.reshape(-1,1), Y1.reshape(-1,1)), axis=1) # Build Arm Feature
    # parameters
    n = x_grid*y_grid # Number of arms
    gamma = 1 / (2 * sig**2)
    K = parallel_kernel(X, gamma)
    return shareable_array(K)

def build_env3_means():
    f = lambda x,y: np.cos(2*np.pi*x)*np.cos(2*np.pi*y)
    x_grid = 30
    y_grid = 30
    x = np.linspace(-1, 1, x_grid)
    y = np.linspace(-1, 1, y_grid)
    X1, Y1 = np.meshgrid(x, y)
    X = np.concatenate((X1.reshape(-1,1), Y1.reshape(-1,1)), axis=1) # Build Arm Feature
    means = np.array([f(X[i][0], X[i][1]) for i in range(x_grid*y_grid)])
    n = 900
    frac = var_dict['frac']
    t = int(np.ceil(n*frac))
    tau = sorted(means, reverse = True)[t]
    # can also just set tau = -100
    return means, tau

def shareable_array(arr):
    '''
    Take in an array arr of doubles. Output a RawArray
    wrapped as an numpy array in the desired shape
    '''
    shape = arr.shape
    A = RawArray('d', int(np.prod(shape)))
    A_np = np.frombuffer(A, dtype=np.float64).reshape(shape)
    np.copyto(A_np, arr)
    del arr
    return A


# build an initializer for the pool
def init_worker(env_func, means_func):
    K = env_func()
    var_dict['K'] = K
    var_dict['K_shape'] = (var_dict['n'], var_dict['n'])
    means,tau = means_func()
    var_dict['means'] = means
    var_dict['tau'] = tau
    
    


# build a worker for each run in the pool
def worker(trial):
    '''Worker function for each parallel task. '''
    np.random.seed(trial)
    n = var_dict['n']
    K = np.frombuffer(var_dict['K']).reshape(var_dict['K_shape'])
    K.setflags(write=False)  # make read-only
    means = var_dict['means']
    tau = var_dict['tau']

    print('running melk trial', trial)
    melk = MELK(means, K, tau,
                noise_sig=var_dict['noise_sig'], 
                gamma=var_dict['gamma'],
                conf_const=var_dict['conf_const'], 
                max_samples=var_dict['max_samples'],
                batch_size=var_dict['batch_size'], 
                verbose=var_dict['verbose'])
    melk.algorithm(step_size=1,
                   design_iters=var_dict['design_iters'], 
                   viz_step=var_dict['design_iters']*2 )
    print('finished melk trial', trial)
    print(' ')

    print('running trial lse', trial)
    alg_lse = LSE(means, K, tau,
                  noise_sig=var_dict['noise_sig'], 
                  conf_const=var_dict['conf_const'], 
                  max_samples=var_dict['max_samples'],
                  record_each=var_dict['batch_size'],
                  verbose=var_dict['verbose'])
    alg_lse.run()
    print('finished trial lse', trial)
    print(' ')
    
    print('running trial truvar', trial)
    alg_truvar = TRUVAR(means, K, tau,
                        noise_sig=var_dict['noise_sig'], 
                        conf_const=var_dict['conf_const'], 
                        max_samples=var_dict['max_samples'],
                        record_each=var_dict['batch_size'],
                        verbose=var_dict['verbose'])
    alg_truvar.run()  
    print('finished trial truvar', trial)

    return (melk, alg_lse, alg_truvar)



if __name__ == '__main__':
    parse_argv()
    # env 1: random draw, env 2: 1 dimensional cosing function, env 3: 2dimensional sin*cos
    # change these to run the associated example
    env_function = build_env1
    means_function = build_env1_means
    with Pool(processes=40,initializer=init_worker,initargs=(env_function, means_function)) as pool:
        all_results = pool.map(worker, range(var_dict['num_trials']))

    # package and save results
    results_melk = [r[0] for r in all_results]
    results_lse = [r[1] for r in all_results]
    results_truvar = [r[2] for r in all_results]
    sig = var_dict['sig']
    noise_sig = var_dict['noise_sig']
    
    with open('results_temp_melk_3_sig_{}_noisesig_{}.pkl'.format(sig, noise_sig),'wb') as f:
        pickle.dump([results_melk], f)
    with open('results_temp_lse_3_sig_{}_noisesig_{}.pkl'.format(sig, noise_sig),'wb') as f:
        pickle.dump([results_lse], f)
    with open('results_temp_truvar_3_sig_{}_noisesig_{}.pkl'.format(sig, noise_sig),'wb') as f:
        pickle.dump([results_truvar], f)


    