import numpy as np 
from scipy.sparse.linalg import eigsh
import conf_intervals as ci
from library import compute_precision, compute_recall
import matplotlib.pyplot as plt

class LSE:
    def __init__(self, means, phiX, K, tau, noise_sig, delta, d_ori=1, nu=None, verbose=False, 
                            kernel_type='linear', bnd=1, max_samples=250000, include_log_factor = True,
                            conf_const=1.96, seed = None, record_each=1000):
        np.random.seed(seed)
        self.phiX = phiX  # points
        self.d = self.phiX.shape[1]
        self.d_ori = d_ori # Dimension of original vectors
        self.n = self.phiX.shape[0]
        self.noise_sig = noise_sig      # std dev of noise
        self.verbose = verbose
        self.G = []     # good arms
        self.B = []      # bad arms
        self.tau = tau      # threshold
        self.sampled = []   # previously sampled arms
        self.means = means  # true means
        self.K = K  # true kernel matrix
        
        self.running_sum = np.zeros(self.d)     # running sum to compute \Phi_t^T @ y_t
        self.delta = delta
        self.max_samples = max_samples
        self.kernel_type = kernel_type
        self.nu = nu        # mattern kernel parameter if
        self.bnd = bnd          # bound on RKHS norm of true function
        self.setup_conf_interval()
        self.arm_count = np.zeros(self.n) # Tracks individual arm pulls
        self.f1_scores = []
        self.precision = []
        self.recall = []
        self.compute_true_set()
        self.sample_points = []
        self.conf_const = conf_const  # constant for confidence width if using straddle
        self.log_factor = include_log_factor
        self.record_each = record_each
        
        
    # def beta_t(self,x):
    #     L = np.max([np.linalg.norm(self.phiX[i]) for i in range(self.n)])
    #     return ci.linear(self.bnd, self.delta, self.d_ori, L, x, self.noise_sig)
        
        

    def compute_true_set(self):
        self.true_set = [i for i in range(self.n) if self.means[i] > self.tau]

    def setup_conf_interval(self):
        if self.kernel_type == 'linear':
            L = np.max([np.linalg.norm(self.phiX[i]) for i in range(self.n)])
            self.beta_t = lambda t: ci.linear(self.bnd, self.delta, self.d_ori, L, t, self.noise_sig)
        elif self.kernel_type == 'rbf':
            self.beta_t = lambda t:  ci.sq_exponential(self.bnd, self.noise_sig, 
                                                       self.delta, t, self.d_ori)
        elif self.kernel_type == 'mattern':
            if self.nu is None: 
                raise Exception('roughness parameter must be specified for Mattern')
            self.beta_t = lambda t: ci.mattern(self.bnd, self.noise_sig, self.delta, t, self.d_ori, self.nu)
        else:
            raise Exception('Invalid choice of kernel_type. Choose from "linear", "rbf", or "mattern"')

    def pull_arm(self, i):
        '''
        Pull single arm i and update everything
        '''
        y = self.means[i] + self.noise_sig * np.random.randn()
        self.running_sum += self.phiX[i] * y        # used for mean computation
        # update covariance and inverse covariance
        self.cov += np.outer(self.phiX[i], self.phiX[i])
        self.inv_cov = LSE.sherman_morrison(self.inv_cov, self.phiX[i])
        self.sampled.append(i)
        # update confidence widths
        self.update_posterior_means()
        self.update_posterior_std()
        self.update_bounds()

    @staticmethod
    def sherman_morrison(Kinv, x):
        '''
        Use Sherman-Morrison to update (K + x x')^-1 given K^-1
        '''
        return Kinv - Kinv @ np.outer(x, x) @ Kinv / (1 + x.T @ Kinv @ x)

    def update_posterior_means(self):
        self.mu = self.phiX @ self.inv_cov @ self.running_sum

    def update_posterior_std(self):
        C = self.cov - self.noise_sig**2 * np.eye(self.d)
        A = self.K - np.diag(self.phiX@self.inv_cov@C@self.phiX.T)
        self.std = np.sqrt(np.abs(A))   # adding abs for numerical stability for small values
        
    def update_bounds(self):
        beta = self.beta_t(len(self.sampled))
        if self.log_factor == False:
            beta = np.sqrt((self.noise_sig**2 + self.bnd**2) * self.conf_const)
        self.upper = self.mu + beta * self.std
        self.lower = self.mu - beta * self.std
        
    def update_good_bad_sets(self):
        for i in range(self.n):
            if i not in self.G and i not in self.B:
                if self.lower[i] > self.tau: self.G.append(i)
                elif self.upper[i] < self.tau: self.B.append(i)

    def compute_ambiguity(self, i):
        '''
        Computes "ambiguity" criterion from Gotovos et al. 
        '''
        return np.maximum(self.upper[i] - self.tau, self.tau - self.lower[i])

    def select_arm(self):
        A = [self.compute_ambiguity(i) if ((i not in self.G) and (i not in self.B))
                                        else -np.inf for i in range(self.n)]
        return np.argmax(A)
            
    def update_metrics(self):
        self.predicted = self.G
        p = compute_precision(self.predicted, self.true_set)
        r = compute_recall(self.predicted, self.true_set)
        self.precision.append(p)
        self.recall.append(r)
        if p == 0 and r == 0: f1 = 0
        else: f1 = 2 * p * r / (p + r)
        self.f1_scores.append(f1)

    def run(self):
        '''
        Run LSE
        '''
        # set everything up
        self.cov = self.noise_sig**2 * np.eye(self.d)
        self.inv_cov = self.noise_sig**(-2) * np.eye(self.d)

        # choose random starting index
        start = np.random.randint(0,self.n-1)
        self.pull_arm(start)
        t = 0
        while len(self.G) + len(self.B) < self.n \
            and len(self.sampled) < self.max_samples:
            arm = self.select_arm()     # choose an arm
            
            self.pull_arm(arm)          # pull that arm
            self.arm_count[arm] += 1    # Count arm pull
            self.update_good_bad_sets()
            t = len(self.sampled)
            if not t % self.record_each: 
                self.update_metrics()
                self.sample_points.append(t)
                
                if self.verbose:
                    print('Samples {}'.format(t))
                    print('{} arms declared good {}'.format(len(self.G), sorted(self.G)))
                    print('{} arms declared bad {}'.format(len(self.B), sorted(self.B)))
                    print('{} arms remaining'.format(self.n - len(self.G) - len(self.B)))
                    plt.plot(self.upper, label = "up")
                    plt.plot(self.lower, label = "dw")
                    plt.plot(self.means, c = "g", label = "m")
                    plt.legend()
                    plt.show()
        print(self.G, t)
        
        return t

class LSE_imp(LSE):
    def __init__(self, means, phiX, K, epsilon, noise_sig, delta, d_ori = 1, nu=None, verbose=False, 
                            kernel_type='linear', bnd=1, max_samples=250000, include_log_factor = True,
                            conf_const=1.96, seed = None, record_each=500):
        super().__init__(means, phiX, K, epsilon, noise_sig, delta, d_ori = d_ori, nu=nu, verbose=verbose, 
                            kernel_type=kernel_type, bnd=bnd, include_log_factor=include_log_factor, 
                             conf_const=conf_const, seed = seed, record_each=record_each)
        self.t_upper = np.inf       # upper bound on threshold
        self.t_lower = -np.inf      # lower bound on threshold
        self.best_lower = -np.inf   # lower bound on highest mean
        self.possible_best = list(range(self.n))     # which arms might be the best
        self.epsilon = epsilon          
        self.max_samples = max_samples
        
        self.tau = (1 - self.epsilon) * max(self.means)
        self.compute_true_set()
        
        

    def compute_true_set(self):
        thresh = self.tau
        self.true_set = [i for i in range(self.n) if self.means[i] > thresh]

    def estimate_threshold(self):
        self.t_upper = (1 - self.epsilon) * max(self.upper[self.possible_best])
        self.best_lower = max(self.lower[self.possible_best])
        self.t_lower = (1 - self.epsilon) * self.best_lower

    def update_good_bad_sets(self):
        self.estimate_threshold()
        for i in range(self.n):
            if i not in self.G and i not in self.B:
                if self.lower[i] > self.t_upper: self.G.append(i)
                elif self.upper[i] < self.t_lower: self.B.append(i)
            if i in self.possible_best:
                if self.upper[i] < self.best_lower: self.possible_best.remove(i)

    def is_active_arm(self, i):
        '''
        Is i and active arm? If not declared bad, or declared good but not best
        '''
        if i in self.B: return False
        elif i in self.G and i not in self.possible_best: return False
        else: return True

    def compute_ambiguity(self, i):
        '''
        Ambiguity defined by width of confidence interval for implicit LSE
        '''
        return self.upper[i] - self.lower[i]

    def select_arm(self):
        '''
        Pull the active arm with the widest confidence intervals
        '''
        A = [self.compute_ambiguity(i) if self.is_active_arm(i)
                                        else -np.inf for i in range(self.n)]
        return np.argmax(A)

    def printout(self, t):
        if self.verbose and not (t % 10000):
            print('Samples {}'.format(t))
            print('{} arms declared good {}'.format(len(self.G), sorted(self.G)))
            print('{} arms declared bad {}'.format(len(self.B), sorted(self.B)))
            print('{} arms remaining'.format(self.n - len(self.G) - len(self.B)))







