import numpy as np
import scipy
from scipy.sparse.linalg import eigsh
import matplotlib.pyplot as plt
import time
from scipy.linalg import lapack


def effective_dim(kernel_func, X, sigma):
    n = X.shape[0]
    K = np.array([[kernel_func(X[i], X[j]) for i in range(n)] for j in range(n)])
    return np.trace(K @ np.linalg.inv(K + sigma * np.eye(n)))

def compute_precision(arr1, arr2):
    '''
    What fraction of elements in arr1 are in arr2
    '''
    if len(arr1) == 0: return 0
    return len(set(arr1).intersection(set(arr2))) / len(arr1)

def compute_recall(arr1, arr2):
    '''
    Compute recall: fraction of elements in arr2 present in arr1
    '''
    if len(arr2) == 0: return 1
    return len(set(arr1).intersection(set(arr2))) / len(arr2)

#Catoni IPS
def catoni_ips(X,V,nu,T,lam,means, delta = .05, reg_l2=0, noise_level=1, 
                                                        return_extra=False,
                                                        past_data=None):
    '''
    Implements algorithm 1 of the RKHS paper using Catoni as a Robust Mean estimator
    input:
        X - design vectors
        V - vectors we want confidence intervals over 
        nu - an upper bound on variance estimate which is just rho
        (TODO: pass in actual variances as opposed to upper bound.)
        T - number of samples
        lam - design over X's
        means - true means
        
    '''
    thetahat_list, armcount = _ips_list(X,T,lam, means, reg_l2=reg_l2, noise_level=noise_level)
    if past_data is not None:
        thetahat_list = np.hstack([past_data, thetahat_list])
    estimates_V = np.zeros((V.shape[0],1))
    for i in range(V.shape[0]):
        thetahat_v_list = V[i] @ thetahat_list
        estimates_V[i] = Catoni(thetahat_v_list,nu*noise_level,delta)
    if not return_extra:
        return estimates_V, armcount
    else:
        theta_hat = np.mean(thetahat_list, axis=1)
        return estimates_V, armcount, theta_hat, thetahat_list
    
# Removing Catoni piece and computing fast IPS via vectorizing
def fast_ips(X,V,nu,T,lam,means, delta = .05, reg_l2=0, noise_level=1, 
                                return_extra=True, past_data=None):
    '''
    Implements algorithm 1 of the RKHS paper using Catoni as a Robust Mean estimator
    input:
        X - design vectors
        V - vectors we want confidence intervals over 
        nu - an upper bound on variance estimate which is just rho
        (TODO: pass in actual variances as opposed to upper bound.)
        T - number of samples
        lam - design over X's
        means - true means
        
    '''
    ### nu, delta is not used in this function ##
    
    thetahat_list, armcount = _ips_list(X,T,lam, means, reg_l2=reg_l2, noise_level=noise_level)
    if past_data is not None:
        thetahat_list = np.hstack([past_data, thetahat_list])
    estimates_V = np.mean(V @ thetahat_list, axis=1)
    if return_extra:
        theta_hat = np.mean(thetahat_list, axis=1)
        return estimates_V, armcount, theta_hat, thetahat_list
    return estimates_V, armcount

def Catoni(Z,nu,delta):
    '''
    input:
        Z is a list of numbers
        nu is estimate of the variance from rho*
        delta confidence level 
    returns:
        Catoni robust estimate of mean of Z
    '''
    n = Z.shape[0]
    alpha = np.sqrt(2*np.log(1/delta)/(n *nu*(1+2*np.log(1/delta)/(n -2*np.log(1/delta)))))

    def test_est(x):
        return np.sum(vec_psi(alpha * (Z - x)))

    mu_hat_low = -1
    mu_hat_high = 1
    #find high and low of the binary search
    if test_est(1) < 0 and test_est(-1) > 0:
        mu_hat_low = -1
        mu_hat_high = 1

    elif test_est(1) > 0:
        mu_hat_high = 1

        while test_est(mu_hat_high) > 0:
            mu_hat_high = mu_hat_high*2

        mu_hat_low = mu_hat_high/2

    elif test_est(-1) < 0:
        mu_hat_high = -1
        while test_est(mu_hat_high) < 0:
            mu_hat_high = mu_hat_high*2

        mu_hat_low = mu_hat_high
        mu_hat_high = mu_hat_high/2

    mu_hat = 0  
    while np.abs(mu_hat_low - mu_hat_high) > 1e-6:
        mu_hat = 1/2*(mu_hat_high + mu_hat_low)
        if test_est(mu_hat) > 0:
            mu_hat_low = mu_hat
        else:
            mu_hat_high = mu_hat        
    return mu_hat
    
def rhostar_mirrordescent(X, Y, reg_l2=0, iters=10000, 
                          step_size=0.01, num_eigs=0, 
                          method='md', viz_step = 10000, initial=None):
    n = X.shape[0]
    d = X.shape[1]
    if initial is not None:
        design = initial
    else:
        design = np.ones(n)
        design /= design.sum()  
    eta = step_size
    grad_norms = []
    history = []
    t = time.time()
    for count in range(1, iters):
        try:
            A_inv = fast_positive_definite_inverse(X.T@np.diag(design)@X + reg_l2*np.eye(d))
        except:
            A_inv = np.linalg.pinv(X.T@np.diag(design)@X + reg_l2*np.eye(d))
#             print(np.linalg.pinv(X.T@np.diag(design)@X + reg_l2*np.eye(d)))
        
        rho = np.diag(Y@A_inv@Y.T)
        y_opt = Y[np.argmax(rho),:]
        g = y_opt @ A_inv @ X.T
        g = -g * g
        if method == 'md':
            eta = step_size / np.sqrt(count + 1)
            design = np.exp(np.log(design) - eta*g/max(np.abs(g)))
            design /= np.sum(design)
        else:
            eta = step_size/(count+2)
            imin = np.argmin(g)
            design = (1-eta)*design+eta*np.eye(n)[imin]
        grad_norms.append(np.linalg.norm(g - np.sum(g)/len(g)*np.ones(n)))
        if count % (viz_step) == 0:
            history.append(np.max(rho))
            print('time per iter', (time.time() - t)/count, count)
            print('rho', rho[np.argmax(rho)])
            fig, ax = plt.subplots(1,2)
            ax[0].plot(grad_norms)
            ax[1].plot(design)
            
            plt.show()
            print('last few grad norms', grad_norms[-10:])
    return design, rho, history


# UTILITY FUNCTIONS
def _ips_list(X,T,lam,means,reg_l2=0, noise_level=1):
    n = X.shape[0]
    d = X.shape[1]
    # speed up with an Einstein Summation
    A_lam = np.sum(np.einsum('i,ij,ik->ijk', lam, X, X), axis=0)
    arm_count = np.zeros(n)
    sampled_arms = np.random.choice(n, size=T, p=lam)
    unique_ids = np.unique(sampled_arms)       # arms with >= 1 pull
    counts = np.bincount(sampled_arms, minlength=n)          # how many times each arm pulled
    # print(counts)
    arm_count[unique_ids] += counts[np.nonzero(counts)]
    try:
        A_lam_inv = fast_positive_definite_inverse(A_lam +reg_l2*np.eye(d))
    except:
        A_lam_inv = np.linalg.pinv(A_lam +reg_l2*np.eye(d))
    rt_vals = means[sampled_arms] + np.random.randn(T) * noise_level 
    thetahat_list = A_lam_inv @ (rt_vals[:, np.newaxis] * X[sampled_arms]).T        
    return thetahat_list, arm_count

#Catoni's Estimator
def _psi(x):
    if x >= 0:
        return np.log(1 +x + np.power(x,2)/2)
    else:
        return -np.log(1-x+np.power(x,2)/2)

vec_psi = np.vectorize(_psi)    # make a vectorized version for speed


def upper_triangular_to_symmetric(ut):
    inds_cache = {}
    n = ut.shape[0]
    try:
        inds = inds_cache[n]
    except KeyError:
        inds = np.tri(n, k=-1, dtype=np.bool)
        inds_cache[n] = inds
    ut[inds] = ut.T[inds]


def fast_positive_definite_inverse(m):
    cholesky, info = lapack.dpotrf(m)
    if info != 0:
        raise ValueError('dpotrf failed on input {}'.format(m))
    inv, info = lapack.dpotri(cholesky)
    if info != 0:
        raise ValueError('dpotri failed on input {}'.format(cholesky))
    upper_triangular_to_symmetric(inv)
    return inv
    