import numpy as np
from numpy import sqrt, log, exp
import itertools
import logging
import time
from library import *
import matplotlib.pyplot as plt 

logging.basicConfig(level= "INFO")

class MELK(object):
    def __init__(self, phiX, tau, means, noise_sig=1.0, delta=.05, reg_l2=1e-3, max_samples=1000000,
                    recycle=True, include_log_factor=True, catoni=False, conf_const = 1.0, viz_step = 1000, seed = None, bnd = 1):
        '''
        Kernel function is type of RKHS kernel to use. If None is given, it is assumed
        to be the linear kernel := k(x,y) = <x,y>
        '''
        
        np.random.seed(seed)
        self.phiX = phiX
        self.M = phiX.shape[0]
        if len(phiX.shape) == 1: self.d = 1
        else: self.d = phiX.shape[1]
        self.tau = tau
        self.means = means
        self.delta = delta
        self.pulls = []
        self.arm_count = np.zeros(self.M) # Tracks individual arm pulls
        self.noise_sig = noise_sig      # std dev of additive noise
        self.reg_l2 = reg_l2
        self.max_samples = max_samples
        self.true_set = [i for i in range(self.M) if self.means[i] > self.tau]
        self.f1_scores = []
        self.precision = []
        self.recall = []
        self.sample_points = []
        self.recycle = recycle  # whether to recycle samples from round to round
        self.past_samples = None   # for recycling data
        self.log_factor = include_log_factor  # whether to include log factor in design
        self.B = bnd      # bound on rewards
        self.catoni = catoni        # whether to use catoni or fast IPS
        self.conf_const = conf_const
        self.viz_step = viz_step
        
    def algorithm(self, seed=None, verbose=False, design_iters=1000, design_step=.01):        
        if seed is not None: np.random.seed(seed)
        self.active_arms = np.arange(self.M, dtype=np.int32)
        self.good_arms = []
        self.arm_counts = np.zeros(self.M)
        self.N = 0
        self.phase_index = 1
        
        while len(self.active_arms) > 0 and self.N <= self.max_samples:                
            logging.info('about to compute allocation')
            # Computing the design and getting estimates
            design, rho, _ = rhostar_mirrordescent(self.phiX, self.phiX[self.active_arms,:],
                                                        iters=design_iters, num_eigs=0, reg_l2=self.reg_l2,
                                                        step_size=design_step, viz_step=self.viz_step, method='fw')
            if self.log_factor: 
                logkz = 2*np.log(2*self.phase_index**2*self.M**2/self.delta) 
            else: logkz = self.conf_const
            num_samples = int(np.ceil(2**(2*self.phase_index)*max(rho)*logkz*(self.noise_sig**2 + self.B**2))) 
            if self.catoni: ips_func = catoni_ips       # do full catoni
            else: ips_func = fast_ips                   # do fast IPS
            self.estimates_X, armcount_phase, self.theta_hat, ips_samples \
                = ips_func(self.phiX, self.phiX[self.active_arms,:], max(rho), num_samples, 
                                                                design, self.means, reg_l2 =self.reg_l2,
                                                                delta=self.delta/(2*self.phase_index**2),
                                                                noise_level=self.noise_sig,
                                                                return_extra=True,
                                                                past_data=self.past_samples)

            if verbose:
                print('logkz {}'.format(logkz))
                print('num_samples {}'.format(num_samples))
            if self.recycle: 
                self.past_samples = ips_samples 
            self.arm_count += armcount_phase 
            self.N += num_samples 
            
            if self.recycle:
                self.compute_upper_lower(rho, self.estimates_X, self.N, logkz)
            else:
                self.compute_upper_lower(rho, self.estimates_X, num_samples, logkz)
            
            if verbose == True:
                plt.plot(self.upper, label = "Up")
                plt.plot(self.lower, label = "Dw")
                plt.legend()
                plt.show()

            if verbose:
                logging.info('round {} total {} rho {}'.format(self.phase_index, num_samples, rho))           
            self.update()
            self.update_emp_means()
            self.update_metrics()
            self.sample_points.append(self.N)
            
            if verbose == True:           
                logging.info('\n\n')
                logging.info('finished phase {}'.format(self.phase_index))
                logging.info('round sample count %s' % str(num_samples))
                logging.info('total sample count %s' % str(self.N))
                logging.info('active arms %s' % str(self.active_arms)) 
                logging.info('good arms %s' % str(self.good_arms)) 
                logging.info('rho %s' % str(rho))     
            self.phase_index += 1
                    
    def compute_upper_lower(self, rho, estimates, T, logkz):
        '''
        rho - variances computed
        estiamtes - IPS or RIPS estimates taken
        T - number of samples drawn
        logkz - log factor from bound
        '''
        m = len(estimates)
        beta = np.sqrt((self.noise_sig**2 + self.B**2) * logkz / T)
        self.upper = np.array([estimates[i] + np.sqrt(rho[i]) * beta for i in range(m)])
        self.lower = np.array([estimates[i] - np.sqrt(rho[i]) * beta for i in range(m)])

    def update(self):         
        ### If this seems slow, then speed up.
        keep_indices = []        
        for i,arm_idx in enumerate(self.active_arms):
            if self.tau > self.upper[i]: 
                pass
            elif self.tau < self.lower[i]:
                self.good_arms.append(arm_idx)
            else:
                keep_indices.append(i)
        self.active_arms = self.active_arms[keep_indices]

    def update_emp_means(self):
        self.emp_means = self.phiX @ self.theta_hat

    def update_metrics(self):
        self.predicted = self.good_arms
        p = compute_precision(self.predicted, self.true_set)
        r = compute_recall(self.predicted, self.true_set)
        self.precision.append(p)
        self.recall.append(r)
        if p == 0 and r == 0: f1 = 0
        else: f1 = 2 * p * r / (p + r)
        self.f1_scores.append(f1)
        
        
        
        
        
