import time
import numpy as np
import itertools
import logging
import time
from library import *

logging.basicConfig(level= "INFO")

class MILK(object):
    def __init__(self, phiX, epsilon, means, noise_sig=1.0, delta=.05, multiplicative=False,
                  reg_l2=1e-2, B = 1, max_samples=15000, include_log_factor=True, recycle=True,
                  catoni=False, linear = False, viz_step = 1000, seed = None, conf_const=1.0):

        self.phiX = phiX
        self.M = phiX.shape[0]
        if len(phiX.shape) == 1: self.d = 1
        else: self.d = phiX.shape[1]
        self.epsilon = epsilon
        self.means = means
        self.delta = delta
        self.noise_sig = noise_sig
        self.pulls = []
        self.mult = multiplicative  #additve or multiplicative epsilon good
        self.arm_count = np.zeros(self.M) # Tracks individual arm pulls
        self.reg_l2 = reg_l2
        self.max_samples = max_samples
        self.emp_means = np.zeros(self.M)
        self.compute_true_set()
        self.f1_scores = []
        self.precision = []
        self.recall = []
        self.sample_points = []
        self.B = B
        self.recycle = recycle  # whether to recycle samples from round to round
        self.past_samples = None   # for recycling data
        self.log_factor = include_log_factor  # whether to include log factor in design
        self.catoni = catoni    # whether to use catoni or fast IPS

        self.linear = linear # Check for linear flag and build Y differently.
        self.viz_step = viz_step
        self.seed = seed
        self.conf_const = conf_const

    def compute_true_set(self):
        '''
        Compute true set of epsilon good arms
        '''
        if self.mult: thresh = (1 - self.epsilon) * max(self.means)
        else: thresh = max(self.means) - self.epsilon
        self.true_eps_good = [i for i in range(self.M) if self.means[i] > thresh]

    def algorithm(self, seed=None, verbose=False, design_iters=5000, design_step=0.01):
        if seed is not None: np.random.seed(seed)
        # self.active_arms = np.arange(self.M_Z, dtype=np.int32)
        self.active_arms = list(range(self.M))

        self.good_arms = [] #Initialize empty good arm set
        self.bad_arms = []  #Initialize empty bad arm set

        self.arm_counts = np.zeros(self.M)
        self.N = 0
        self.phase_index = 1

        self.build_Y() # Build Y = x' - x or x - (1-epsilon)x'
        self.A_set = self.Yhat.copy()
        self.A_idxs = self.idxs.copy() # Contains the arm id's of x' - x
        self.A_idxs = self.A_idxs.astype(int) # Make sure the indices are integers
        self.A_idxs_flatten = np.array([i for i in range(len(self.A_idxs))])


        while len(self.good_arms) + len(self.bad_arms) < self.M and \
                    self.N <= self.max_samples: #Stopping condition
            self.delta_t = self.delta/(self.phase_index**2)
            # ts = time.time()
            design, rho, hist = rhostar_mirrordescent(self.phiX,self.Yhat[self.A_idxs_flatten,:],
                                                    iters=design_iters, step_size=design_step,
                                                    reg_l2=self.reg_l2, num_eigs=0, method='fw', viz_step = self.viz_step)
            if self.log_factor:
                logkz = np.log(2 * self.M**2 / self.delta_t)
            else: logkz = self.conf_const
            num_samples = int(np.ceil(2**(2*self.phase_index)*max(rho)*logkz*(self.noise_sig**2 + self.B**2)))
            if self.catoni: ips_func = catoni_ips       # do full catoni
            else: ips_func = fast_ips                   # do fast IPS
            self.X_estimates, armcount_phase, self.theta_hat, ips_samples = ips_func(self.phiX,
                                                                        self.Yhat[self.A_idxs_flatten,:],
                                                                        max(rho), num_samples, design, self.means,
                                                                        noise_level=self.noise_sig,
                                                                        return_extra=True,
                                                                        past_data=self.past_samples)

            self.arm_count += armcount_phase
            if self.recycle:
                self.past_samples = ips_samples
            self.N += num_samples # Count total pulls

            self.compute_upper_lower(rho, self.X_estimates, self.N, logkz)
            self.drop_bad_arms()  # Drop Bad arms
            self.add_good_arms()  # Add Good arms

            self.update_emp_means()
            self.update_metrics()
            self.sample_points.append(self.N)

            logging.info('round {} total {} rho {}'.format(self.phase_index, num_samples, max(rho)))


            if verbose == True:
                logging.info('\n\n')
                logging.info('finished phase {}'.format(self.phase_index))
                logging.info('round sample count %s' % str(num_samples))
                logging.info('total sample count %s' % str(self.N))
                logging.info('active arms %s' % str(self.active_arms))
                logging.info('good arms %s' % str(self.good_arms))
                # logging.info('rho %s' % str(rho))
                plt.plot(self.upper, label = "Up")
                plt.plot(self.lower, label = "Dw")
                plt.legend()
                plt.show()


            self.phase_index += 1

    def compute_upper_lower(self, rho, estimates, T, logkz):
        '''
        rho - variances computed
        estiamtes - IPS or RIPS estimates taken
        T - number of samples drawn
        logkz - log factor from bound
        '''
        m = len(estimates)
        beta = np.sqrt((self.noise_sig**2 + self.B**2) * logkz / T)
        self.upper = np.array([estimates[i] + np.sqrt(rho[i]) * beta for i in range(m)])
        self.lower = np.array([estimates[i] - np.sqrt(rho[i]) * beta for i in range(m)])

    def update_emp_means(self):
        self.emp_means = self.phiX @ self.theta_hat

    def update_metrics(self):
        if self.mult:
            thresh = (1 - self.epsilon) * max(self.emp_means)
        else:
            thresh = max(self.emp_means) - self.epsilon
        self.predicted = self.good_arms
        p = compute_precision(self.predicted, self.true_eps_good)
        r = compute_recall(self.predicted, self.true_eps_good)
        self.precision.append(p)
        self.recall.append(r)
        if p == 0 and r == 0: f1 = 0
        else: f1 = 2 * p * r / (p + r)
        self.f1_scores.append(f1)

    def build_Y(self):
        '''
        build all difference vectors of the Z's.
        '''
        k = self.M     # for brevity
        idxs = np.zeros((k**2,2))

        if self.linear == True: # k^2 by d in Linear case
            Y = np.zeros((k**2, self.d))
        else:
            Y = np.zeros((k**2, k))    # k^2 by k in RKHS case
        for i in range(k):
            idxs[k*i:k*(i+1),0] = np.arange(k)
            idxs[k*i:k*(i+1),1] = i
            if self.mult:
                Y[k*i:k*(i+1),:] = self.phiX[i,:] - (1 - self.epsilon) * self.phiX
            else:
                Y[k*i:k*(i+1),:] = self.phiX - self.phiX[i,:]
        self.Yhat = Y
        self.idxs = idxs

    def drop_bad_arms(self):
        A_set = self.A_set.copy() # Copy to keep original A_set unchanged
        A_idxs = self.A_idxs.copy()
        A_idxs_flatten = self.A_idxs_flatten.copy()

        removeidxs = []
        for i in range(len(A_set)):
            if i in removeidxs:
                continue
            # remove bad arm and all differences
            if (not self.mult and self.lower[i] > self.epsilon) or \
                (self.mult and self.upper[i] < 0):
                arm_idx = A_idxs[i, 1]
                if arm_idx in self.active_arms:
                    self.active_arms.remove(arm_idx)  # remove bad arms from active set
                if int(arm_idx) not in self.bad_arms:
                    self.bad_arms.append(int(arm_idx)) # add bad arms to B

                removeidxs += list(np.where(A_idxs[:,1] == arm_idx)[0]) # Add indices to the remove list
                removeidxs += list(np.where(A_idxs[:,0] == arm_idx)[0]) # additional bad arm removal

            elif (not self.mult and self.upper[i] < self.epsilon) or \
                  (self.mult and self.lower[i] > 0):
                if i in removeidxs: continue
                removeidxs += [i]
        self.A_set = np.delete(A_set, list(set(removeidxs)), axis=0)
        self.A_idxs = np.delete(A_idxs, list(set(removeidxs)), axis=0).astype(int)
        self.A_idxs_flatten = np.delete(A_idxs_flatten, list(set(removeidxs)), axis=0).astype(int)

    def add_good_arms(self,):
        for i in self.active_arms:
            a = list(np.where(self.A_idxs[:,1] == i)[0])
            if len(a) == 0:
                if i not in self.good_arms:
                    self.good_arms.append(i)
        return self.good_arms
